'use strict';

exports.__esModule = true;

var _toughCookie = require('tough-cookie');

var _cookieLimit = require('./cookie-limit');

var _cookieLimit2 = _interopRequireDefault(_cookieLimit);

var _lodash = require('lodash');

var _url = require('../utils/url');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var LOCALHOST_DOMAIN = 'localhost';
var LOCALHOST_IP = '127.0.0.1';

var Cookies = function () {
    function Cookies() {
        _classCallCheck(this, Cookies);

        this.cookieJar = new _toughCookie.CookieJar();
    }

    Cookies._hasLocalhostDomain = function _hasLocalhostDomain(cookie) {
        if (cookie) return cookie.domain === LOCALHOST_DOMAIN || cookie.domain === LOCALHOST_IP;

        return false;
    };

    Cookies.prototype._set = function _set(url, cookies, isClient) {
        var _this = this;

        cookies = (0, _lodash.castArray)(cookies);

        return cookies.reduce(function (resultCookies, cookieStr) {
            if (cookieStr.length > _cookieLimit2.default) return resultCookies;

            var cookie = _toughCookie.Cookie.parse(cookieStr);

            // NOTE: If cookie.domain and url hostname are equal to localhost/127.0.0.1,
            // we should remove 'Domain=...' form cookieStr (GH-1491)
            if (Cookies._hasLocalhostDomain(cookie) && (0, _url.parseUrl)(url).hostname === cookie.domain) {
                cookie.domain = '';
                cookieStr = cookie.toString();
            }

            var parsedCookie = _this.cookieJar.setCookieSync(cookieStr, url, {
                http: !isClient,
                ignoreError: true,
                loose: true
            });

            if (parsedCookie) resultCookies.push(parsedCookie);

            return resultCookies;
        }, []);
    };

    Cookies.prototype.serializeJar = function serializeJar() {
        return JSON.stringify(this.cookieJar.serializeSync());
    };

    Cookies.prototype.setJar = function setJar(serializedJar) {
        this.cookieJar = serializedJar ? _toughCookie.CookieJar.deserializeSync(JSON.parse(serializedJar)) : new _toughCookie.CookieJar();
    };

    Cookies.prototype.setByServer = function setByServer(url, cookies) {
        return this._set(url, cookies, false);
    };

    Cookies.prototype.setByClient = function setByClient(url, cookies) {
        this._set(url, cookies, true);
    };

    Cookies.prototype.getClientString = function getClientString(url) {
        return this.cookieJar.getCookieStringSync(url, { http: false });
    };

    Cookies.prototype.getHeader = function getHeader(url) {
        return this.cookieJar.getCookieStringSync(url, { http: true }) || null;
    };

    return Cookies;
}();

exports.default = Cookies;
module.exports = exports['default'];